from pyg_base import timer
from pyg_cell import get_DAG, add_edge, del_edge, topological_sort, descendants

from pyg_base._dag import get_DAGs
import networkx as nx


def test_get_DAG():
    DAGs = get_DAGs()
    DAGs[None] = {}
    dag = get_DAG()
    assert len(dag) == 0
    dag = add_edge('a', 'b')
    assert len(get_DAG()) == 1
    dag = add_edge('a', 'c')
    dag = add_edge('b', 'c', dag)
    assert len(get_DAG()) == 2
    assert len(get_DAG()['a']) == 2
    dag = del_edge('a', 'c')
    assert len(get_DAG()['a']) == 1
    assert topological_sort(None, 'a')['node2gen'] == {'a': 0, 'b': 1, 'c': 2}
    assert descendants(None, 'a') == ['a', 'b', 'c']
    assert descendants(None, 'b') == ['b', 'c']
    assert descendants(None, 'a', exc = 0) == ['b', 'c']



def dag2G(dag):
    G = nx.DiGraph()
    for a in dag:
        for b in dag[a]:
            G.add_edge(a,b)
    return G


def compare_topological_sort_and_nx(gen2node,y):
    if 'gen2node' in gen2node:
        gen2node = gen2node['gen2node']
    i = 1    
    for g, nodes in sorted(gen2node.items())[1:]:
        n = len(nodes)            
        if sorted(y[i:i+n]) != sorted(nodes):
            return False
        i+=n
    return True


def _nx_topological_sort(G, nodes):
   return nx.algorithms.dag.topological_sort(G.subgraph(nodes))
   
    
def nx_topological_sort(dag, node):
   G = dag2G(dag)
   nodes = {node} | nx.descendants(G, node)
   return list(_nx_topological_sort(G, nodes))

def test_topological_sort():
    dag = dict(a = dict(b = 0, c = 0), b = dict(c = 0, d = 0, e = 0), c = dict(e = 0), d = dict(e = 0))
    node = 'a'
    assert topological_sort(dag, node)['node2gen'] == {'a': 0, 'b': 1, 'c': 2, 'e': 3, 'd': 2}
    x = topological_sort(dag, node)
    y = nx_topological_sort(dag, node)
    assert compare_topological_sort_and_nx(x,y)

    
    get_DAGs()['test'] = {}
    dag = get_DAG('test')
    for b in range(100):
        for a in range(b-3,b+3):
            add_edge('a%i'%a, 'b%i'%b, dag)
    for c in range(100):
        for a in range(c-3,c+3):
            add_edge('a%i'%a, 'c%i'%c, dag)
        for b in range(c-3,c+3):
            add_edge('b%i'%b, 'c%i'%c, dag)
    for d in range(100):
        for c in range(d-10,d+10):
            add_edge('c%i'%c, 'd%i'%d, dag)
    for e in range(100):
        for c in range(e-10,e+10):
            add_edge('c%i'%c, 'e%i'%e, dag)
        for a in range(e-3,e+3):
            add_edge('a%i'%a, 'e%i'%e, dag)

    for node in ['a50', 'a10', 'b40', 'c13']:
        x = topological_sort(dag, node)
        y = nx_topological_sort(dag, node)
        assert compare_topological_sort_and_nx(x,y)

    G = dag2G(dag)
    nodes = {node} | nx.descendants(G, node)
    t0 = timer(topological_sort, n = 1000, time = True)(dag, node)
    t1 = timer(_nx_topological_sort, n = 1000, time = True)(G, nodes)
    assert t1 / t0 





from pyg_base import dt, dt_bump, calendar, cfg_read, is_date, as_list
from pyg_cell._db_cell import db_cell, _updated

__all__ = ['periodic_cell']

_period = 'period'

_day_start = cfg_read().get('day_start', 0)
_day_end = cfg_read().get('day_end', 235959)
_end_date = 'end_date'
END_DATE = ['maturity', 'last_tradeable_dt', 'opt_expire_dt']

class periodic_cell(db_cell):
    """
    periodic_cell inherits from db_cell its ability to save itself in MongoDb using its db members
    Its calculation schedule depends on when it was last updated. There are two "reserved" keys associated with periodic_cell

    1) period   : determine frequency of update schedule
    2) end_date : determine when to stop updating
    
    In addition, there are three more parameters used in determining business days, if period is in business days

    3) calendar : determines the calendar used 
    4) day_start: time when business day starts
    5) day_end  : time when business day ends
    
    :Example: periodic cell and 'period' parameter
    ---------
    >>> from pyg import *
    >>> c = periodic_cell(lambda a: a + 1, a = 0)
    >>> assert c.period == '1b'
    
    default behaviour for period is 1 business day

    We now assert it needs to be calculated and calculate it...
    
    >>> assert c.run()
    >>> c = c.go()
    >>> assert c.data == 1
    >>> assert not c.run()
    
    Now let us cheat and tell it, it was last run 3 days ago...
    
    >>> c.updated = dt(-3)
    >>> assert c.run()

    Now we change the update frequency to '1w'
    >>> c.period = '1w'
    >>> assert not c.run() # since I ran three days ago already
    
    Example: periodic_cell and 'end_date'
    --------
    'end_date' is a reserved key for a periodic cell determining when it "expires"
    Alternatively, end_date is a list of keys, each of which can contain an expiry date.
    By default, ['maturity', 'last_tradeable_dt', 'opt_expire_dt'] all designate expiries
    
    All the above cells will stop updating past 2022/01/01:
    >>> c = periodic_cell(some_func, end_date = dt(2022)) 
    >>> c = periodic_cell(some_func, end_date = 'expiry', expiry = 2022) #end_date redirects to "expiry"
    >>> c = periodic_cell(some_func, maturity = 2022) #end_date redirects to maturity as a default
    
    """
    def __init__(self, function = None, output = None, db = None, period = '1b', updated = None, **kwargs):
        self[_updated] = updated
        self[_period] = period 
        super(periodic_cell, self).__init__(function, output = output, db = db, **kwargs)
            
    def run(self):        
        end_date = self.get(_end_date, END_DATE)
        time = dt()
        updated = self.get(_updated)
        # if self[_updated] is None or dt_bump(self[_updated], self[_period]) < time:
        if is_date(end_date):
            edate = end_date
            if time > dt(edate, 1) and updated is not None and dt_bump(updated, self[_period]) >= edate: ## we expired and last ran near expiry
                    return super(periodic_cell, self).run() 
        else:
            for e in as_list(end_date):
                edate = self.get(e)
                if edate: # exclude 0 and None
                    edate = dt(edate)                
                    if time > dt(edate, 1) and updated is not None and dt_bump(updated, self[_period]) >= edate: ## we expired and last ran near expiry
                        return super(periodic_cell, self).run() 
        if updated is None:
            return True
        else:
            cal = calendar(self.get('calendar'))
            next_update = cal.dt_bump(updated, self.period)
            day_start = self.get('day_start', _day_start)
            day_end = self.get('day_end', _day_end)                            
            if cal.trade_date(time, 'p', day_start, day_end) >= cal.trade_date(next_update, 'f', day_start, day_end):
                return True
            else:
                return super(periodic_cell, self).run() 


    



from pyg_base import dictable, loop, add_, getargspec
from pyg_cell import cell, cell_runner# -*- coding: utf-8 -*-


def test_cell_runner_decorate():
    W = cell_runner(decorate = loop(dict), as_data = True)
    rs = dictable(a = [dict(x = 1, y = 2), 
                           3, 
                           dict(x = 3, y = 4)], 
                      b = [5, 6, dict(x = 1, y = 2)],
                      name = ['a is dict', 'neither dict', 'both dicts'])
    rs = rs(c = W(lambda a, b: cell(add_, a = a, b = b)))
    assert rs.c == [dict(x = 6, y = 7), 9, dict(x = 4, y = 6)]


def test_cell_runner_changes_signature():
    ADDS_DB_KEYS_TO_ARGSPEC= cell_runner(db = ['x', 'key'])(lambda a, b: a + b)
    assert getargspec(ADDS_DB_KEYS_TO_ARGSPEC).args == ['x', 'key', 'a', 'b']    
    REPLACES_SPEC_USING_ARGS = cell_runner(args = dict(key = 'item'), db = ['x', 'item'])(lambda a, b: a+b)
    assert sorted(getargspec(REPLACES_SPEC_USING_ARGS).args) == ['a', 'b', 'key', 'x']



_db = 'db'
_root = 'root'
_writer = 'writer'
_table = 'table'
from pyg_base import dictattr, tree_setitem, get_cache, tree_repr
from pyg_encoders import root_path

def _doc_writer(doc):
    if _root in doc:
        return doc[_root]
    elif _db in doc:
        db = doc[_db]
        return db.keywords.get(_writer)
 
def _address_key(address):
    k,v = zip(*address)
    return k

def _doc_key(doc, address = None):
    address = address or doc._address
    if _db in doc:
        pk = doc._pk
        return tuple([pair[0] if pair[0] in pk else pair[1] for pair in address])
    else:
        return _address_key(doc._address)

def _doc_write(doc, writer, tree):
    path = root_path(doc, writer).lower().replace(':', '').replace(' ', '_').replace('+','_')
    nodes = [p.split('?')[0] for p in path.split('/') if len(p)]
    nodes[-1] = '.'.join(nodes[-1].split('.')[:-1])
    tree_setitem(tree, nodes, doc)
    

def _trim_all(tree):
    if isinstance(tree, dict):
        if len(tree) == 1:
            return _trim_all(list(tree.values())[0])
        else:
            return type(tree)({k: _trim_all(v) for k,v in tree.items()})
    else:
        return tree
    
def cell_tree(docs = None, tree = None, skip = True, mappers = None, trim = 'all'):
    ## This creates a graph based on the document primary keys
    """
    from pyg import  * 
    ticker = mongo_table('ticker', 'momentum')
    tdocs = ticker.inc(item = ['buffered_signal', 'uncertainty', 'combined_signal']).docs()
    tdocs = tdocs.do(cell_go)
    
    predictor = mongo_table('predictor', 'momentum')
    pdocs = predictor.inc(predictor = ['mom_1', 'mom_2']).docs()
    
    >>> tree = cell_tree()
    >>> tree = cell_tree(mappers = {('localhost:27017', 'cme', 'expiry', 'ticker'): 'C:/Users/Dell/Dropbox/Yoav/TradingData/bbg/ticker/%ticker/%expiry.parquet', 
                                    ('localhost:27017', 'ice', 'expiry', 'ticker'): 'C:/Users/Dell/Dropbox/Yoav/TradingData/bbg/ticker/%ticker/%expiry.parquet'})
    >>> tree = cell_tree(docs.doc + pdocs.doc + [cell(pk = ['a'], a = 1)])
    """
    missing = {}
    really_missing = {}
    mappers = mappers or {}
    writers = {}
    if tree is None:
        tree = dictattr()
    if docs is None:
        docs = get_cache()['GRAPH']
    if isinstance(docs, list):
        docs = {doc._address : doc for doc in docs}
    for address, doc in docs.items():
        pass
        key = _doc_key(doc, address)
        if key in mappers:
            writer = mappers[key]
        else:
            writer = _doc_writer(doc)
            if writer is not None:
                writers[key] = writer
            elif key in writers:
                writer = writers[key]
            else:
                missing[address] = doc
        if writer is not None:
            _doc_write(doc, writer, tree)
    for address, doc in missing.items():
        key = _doc_key(doc, address)
        if key not in writers:
            if not skip:
                raise ValueError('not sure how to add this key %s'%key)
            else:
                really_missing[key] = address 
        else:
            writer = writers[key]
            _doc_write(doc, writer, tree)
    if trim == 'leading':
        while isinstance(tree, dict) and len(tree) == 1:
            tree = list(tree.values())[0]
    elif trim == 'all':
        tree = _trim_all(tree)
    if really_missing:
        print('we were not able to map these keys into a logical place within the graph', really_missing.keys())
        print('these got mapped:\n', tree_repr(writers))
    return tree
                
            

    
    
    
    
    
    